@@ -281,7 +281,8 @@ def prepare_request_dict(request_dict, endpoint_url, context=None,
     if user_agent is not None:
         headers = r['headers']
         headers['User-Agent'] = user_agent
-    url = _urljoin(endpoint_url, r['url_path'])
+    host_prefix = r.get('host_prefix')
+    url = _urljoin(endpoint_url, r['url_path'], host_prefix)
     if r['query_string']:
         # NOTE: This is to avoid circular import with utils. This is being
         # done to avoid moving classes to different modules as to not cause
@@ -318,7 +319,7 @@ def create_request_object(request_dict):
     return request_object
 
 
-def _urljoin(endpoint_url, url_path):
+def _urljoin(endpoint_url, url_path, host_prefix):
     p = urlsplit(endpoint_url)
     # <part>   - <index>
     # scheme   - p[0]
@@ -330,24 +331,132 @@ def _urljoin(endpoint_url, url_path):
         # If there's no path component, ensure the URL ends with
         # a '/' for backwards compatibility.
         if not p[2]:
-            return endpoint_url + '/'
-        return endpoint_url
-    if p[2].endswith('/') and url_path.startswith('/'):
+            new_path = '/'
+        else:
+            new_path = p[2]
+    elif p[2].endswith('/') and url_path.startswith('/'):
         new_path = p[2][:-1] + url_path
     else:
         new_path = p[2] + url_path
-    reconstructed = urlunsplit((p[0], p[1], new_path, p[3], p[4]))
+
+    new_netloc = p[1]
+    if host_prefix is not None:
+        new_netloc = host_prefix + new_netloc
+
+    reconstructed = urlunsplit((p[0], new_netloc, new_path, p[3], p[4]))
     return reconstructed
 
 
+class AWSRequestPreparer(object):
+    """
+    This class performs preparation on AWSRequest objects similar to that of
+    the PreparedRequest class does in the requests library. However, the logic
+    has been boiled down to meet the specific use cases in botocore. Of note
+    there are the following differences:
+        This class does not heavily prepare the URL. Requests performed many
+        validations and corrections to ensure the URL is properly formatted.
+        Botocore either performs these validations elsewhere or otherwise
+        consistently provides well formatted URLs.
+
+        This class does not heavily prepare the body. Body preperation is
+        simple and supports only the cases that we document: bytes and
+        file-like objects to determine the content-length. This will also
+        additionally prepare a body that is a dict to be url encoded params
+        string as some signers rely on this. Finally, this class does not
+        support multipart file uploads.
+
+        This class does not prepare the method, auth or cookies.
+    """
+    def prepare(self, original):
+        method = original.method
+        url = self._prepare_url(original)
+        body = self._prepare_body(original)
+        headers = self._prepare_headers(original, body)
+        stream_output = original.stream_output
+
+        return AWSPreparedRequest(method, url, headers, body, stream_output)
+
+    def _prepare_url(self, original):
+        url = original.url
+        if original.params:
+            params = urlencode(list(original.params.items()), doseq=True)
+            url = '%s?%s' % (url, params)
+        return url
+
+    def _prepare_headers(self, original, prepared_body=None):
+        headers = HeadersDict(original.headers.items())
+
+        # If the transfer encoding or content length is already set, use that
+        if 'Transfer-Encoding' in headers or 'Content-Length' in headers:
+            return headers
+
+        # Ensure we set the content length when it is expected
+        if original.method not in ('GET', 'HEAD', 'OPTIONS'):
+            length = self._determine_content_length(prepared_body)
+            if length is not None:
+                headers['Content-Length'] = str(length)
+            else:
+                # Failed to determine content length, using chunked
+                # NOTE: This shouldn't ever happen in practice
+                body_type = type(prepared_body)
+                logger.debug('Failed to determine length of %s', body_type)
+                headers['Transfer-Encoding'] = 'chunked'
+
+        return headers
+
+    def _to_utf8(self, item):
+        key, value = item
+        if isinstance(key, six.text_type):
+            key = key.encode('utf-8')
+        if isinstance(value, six.text_type):
+            value = value.encode('utf-8')
+        return key, value
+
+    def _prepare_body(self, original):
+        """Prepares the given HTTP body data."""
+        body = original.data
+        if body == b'':
+            body = None
+
+        if isinstance(body, dict):
+            params = [self._to_utf8(item) for item in body.items()]
+            body = urlencode(params, doseq=True)
+
+        return body
+
+    def _determine_content_length(self, body):
+        # No body, content length of 0
+        if not body:
+            return 0
+
+        # Try asking the body for it's length
+        try:
+            return len(body)
+        except (AttributeError, TypeError) as e:
+            pass
+
+        # Try getting the length from a seekable stream
+        if hasattr(body, 'seek') and hasattr(body, 'tell'):
+            orig_pos = body.tell()
+            body.seek(0, 2)
+            end_file_pos = body.tell()
+            body.seek(orig_pos)
+            return end_file_pos - orig_pos
+
+        # Failed to determine the length
+        return None
+
+
 class AWSRequest(object):
     """Represents the elements of an HTTP request.
 
-    This class is originally inspired by requests.models.Request, but has been
+    This class was originally inspired by requests.models.Request, but has been
     boiled down to meet the specific use cases in botocore. That being said this
     class (even in requests) is effectively a named-tuple.
     """
 
+    _REQUEST_PREPARER_CLS = AWSRequestPreparer
+
     def __init__(self,
                  method=None,
                  url=None,
@@ -357,6 +466,8 @@ class AWSRequest(object):
                  auth_path=None,
                  stream_output=False):
 
+        self._request_preparer = self._REQUEST_PREPARER_CLS()
+
         # Default empty dicts for dict params.
         params = {} if params is None else params
 
@@ -383,51 +494,34 @@ class AWSRequest(object):
 
     def prepare(self):
         """Constructs a :class:`AWSPreparedRequest <AWSPreparedRequest>`."""
-        return AWSPreparedRequest(self)
+        return self._request_preparer.prepare(self)
 
     @property
     def body(self):
-        p = AWSPreparedRequest(self)
-        p.prepare_body(self.data)
-        if isinstance(p.body, six.text_type):
-            p.body = p.body.encode('utf-8')
-        return p.body
+        body = self.prepare().body
+        if isinstance(body, six.text_type):
+            body = body.encode('utf-8')
+        return body
 
 
 class AWSPreparedRequest(object):
-    """Represents a prepared request.
+    """A data class representing a finalized request to be sent over the wire.
 
-    This class is originally inspired by requests.models.PreparedRequest, but
-    has been boiled down to meet the specific use cases in botocore. Of note
-    there are the following differences:
-        This class does not heavily prepare the URL. Requests performed many
-        validations and corrections to ensure the URL is properly formatted.
-        Botocore either performs these validations elsewhere or otherwise
-        consistently provides well formatted URLs.
+    Requests at this stage should be treated as final, and the properties of
+    the request should not be modified.
 
-        This class does not heavily prepare the body. Body preperation is
-        simple and supports only the cases that we document: bytes and
-        file-like objects to determine the content-length. This will also
-        additionally prepare a body that is a dict to be url encoded params
-        string as some signers rely on this. Finally, this class does not
-        support multipart file uploads.
-
-        This class does not prepare the method, auth or cookies.
-
-    :ivar method: HTTP Method
+    :ivar method: The HTTP Method
     :ivar url: The full url
     :ivar headers: The HTTP headers to send.
     :ivar body: The HTTP body.
     :ivar stream_output: If the response for this request should be streamed.
-    :ivar original: The original AWSRequest
     """
-    def __init__(self, original):
-        self.method = original.method
-        self.prepare_url(original.url, original.params)
-        self.prepare_headers(original.headers)
-        self.prepare_body(original.data)
-        self.stream_output = original.stream_output
-        self.original = original
+    def __init__(self, method, url, headers, body, stream_output):
+        self.method = method
+        self.url = url
+        self.headers = headers
+        self.body = body
+        self.stream_output = stream_output
 
     def __repr__(self):
         fmt = (
@@ -437,6 +531,12 @@ class AWSPreparedRequest(object):
         return fmt % (self.stream_output, self.method, self.url, self.headers)
 
     def reset_stream(self):
+        """Resets the streaming body to it's initial position.
+
+        If the request contains a streaming body (a streamable file-like object)
+        seek to the object's initial position to ensure the entire contents of
+        the object is sent. This is a no-op for static bytes-like body types.
+        """
         # Trying to reset a stream when there is a no stream will
         # just immediately return.  It's not an error, it will produce
         # the same result as if we had actually reset the stream (we'll send
@@ -453,69 +553,18 @@ class AWSPreparedRequest(object):
             logger.debug("Unable to rewind stream: %s", e)
             raise UnseekableStreamError(stream_object=self.body)
 
-    def prepare_url(self, url, params):
-        if params:
-            params = urlencode(list(params.items()), doseq=True)
-            self.url = '%s?%s' % (url, params)
-        else:
-            self.url = url
-
-    def prepare_headers(self, headers):
-        headers = headers or {}
-        self.headers = HeadersDict(headers.items())
-
-    def _to_utf8(self, item):
-        key, value = item
-        if isinstance(key, six.text_type):
-            key = key.encode('utf-8')
-        if isinstance(value, six.text_type):
-            value = value.encode('utf-8')
-
-        return key, value
-
-    def prepare_body(self, data):
-        """Prepares the given HTTP body data."""
-        self.body = data
-
-        if self.body == b'':
-            self.body = None
-
-        if not self.body and self.method == 'GET':
-            return
-
-        if self.body is None:
-            self.headers['Content-Length'] = '0'
-
-        if isinstance(self.body, dict):
-            params = [self._to_utf8(item) for item in self.body.items()]
-            self.body = urlencode(params, doseq=True)
-
-        try:
-            length = len(self.body)
-            self.headers['Content-Length'] = str(length)
-        except (AttributeError, TypeError) as e:
-            pass
-
-        if 'Content-Length' not in self.headers:
-            if hasattr(data, 'seek') and hasattr(data, 'tell'):
-                orig_pos = data.tell()
-                data.seek(0, 2)
-                end_file_pos = data.tell()
-                self.headers['Content-Length'] = str(end_file_pos - orig_pos)
-                data.seek(orig_pos)
-
-        if self.body and 'Content-Length' not in self.headers:
-            # NOTE: This should probably never happen, we don't use chunked
-            self.headers['Transfer-Encoding'] = 'chunked'
-
 
 class AWSResponse(object):
-    """
-    This class is originally inspired by requests.models.Response, but
-    has been boiled down to meet the specific use cases in botocore. This
-    has effectively been reduced to a named tuple for our use case. Most of
-    the more interesting functionality from the requests version has been
-    put onto our botocore.response.StreamingBody class.
+    """A data class representing an HTTP response.
+
+    This class was originally inspired by requests.models.Response, but has
+    been boiled down to meet the specific use cases in botocore. This has
+    effectively been reduced to a named tuple.
+
+    :ivar url: The full url.
+    :ivar status_code: The status code of the HTTP response.
+    :ivar headers: The HTTP headers received.
+    :ivar body: The HTTP response body.
     """
 
     def __init__(self, url, status_code, headers, raw):
@@ -528,7 +577,7 @@ class AWSResponse(object):
 
     @property
     def content(self):
-        """Content of the response, in bytes."""
+        """Content of the response as bytes."""
 
         if self._content is None:
             # Read the contents.
@@ -541,6 +590,12 @@ class AWSResponse(object):
 
     @property
     def text(self):
+        """Content of the response as a proper text type.
+
+        Uses the encoding type provided in the reponse headers to decode the
+        response content into a proper text type. If the encoding is not
+        present in the headers, UTF-8 is used as a default.
+        """
         encoding = botocore.utils.get_encoding_from_headers(self.headers)
         if encoding:
             return self.content.decode(encoding)
